@@ -7,29 +7,35 @@ use crate::{
     symbol::OptionSymbol,
 };
 
-use chrono::{DateTime, FixedOffset};
+use chrono::{DateTime, FixedOffset, NaiveDate};
 use num_rational::Rational64;
 use num_traits::{Signed, Zero};
 use serde::{Deserialize, Serialize};
 
-#[derive(Debug, Serialize, Deserialize)]
+use std::cmp::Ordering;
+
+#[derive(Clone, Debug, Serialize, Deserialize)]
 #[serde(rename_all = "kebab-case")]
 pub(crate) struct Response<Data> {
     pub data: Data,
     pub pagination: Option<Pagination>,
 }
 
-#[derive(Debug, Serialize, Deserialize)]
+#[derive(Clone, Debug, Serialize, Deserialize)]
 #[serde(rename_all = "kebab-case")]
 pub struct Pagination {
     pub page_offset: i32,
     pub total_pages: i32,
 }
 
-#[derive(Debug, Eq, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Copy, Debug, Eq, PartialEq, Serialize, Deserialize, Hash)]
 pub enum InstrumentType {
-    Future,
     Equity,
+    #[serde(rename = "Equity Option")]
+    EquityOption,
+    Future,
+    #[serde(rename = "Future Option")]
+    FutureOption,
     Index,
     Cryptocurrency,
     Unknown,
@@ -38,17 +44,17 @@ pub enum InstrumentType {
 pub mod accounts {
     use super::*;
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub(crate) struct Response {
         pub items: Vec<Item>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub struct Item {
         pub account: Account,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct Account {
         pub account_number: String,
@@ -58,19 +64,19 @@ pub mod accounts {
 pub mod watchlists {
     use super::*;
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub(crate) struct Response {
         pub items: Vec<Item>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub struct Item {
         pub name: String,
         #[serde(rename = "watchlist-entries")]
         pub entries: Vec<Entry>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub struct Entry {
         pub symbol: String,
         #[serde(alias = "instrument-type")] // appears as both kebab and snake case
@@ -81,12 +87,12 @@ pub mod watchlists {
 pub mod market_metrics {
     use super::*;
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub(crate) struct Response {
         pub items: Vec<Item>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct Item {
         pub symbol: String,
@@ -115,9 +121,10 @@ pub mod market_metrics {
         #[serde(with = "string_serialize")]
         pub updated_at: DateTime<FixedOffset>,
         pub option_expiration_implied_volatilities: Option<Vec<ExpirationImpliedVolatility>>,
+        pub earnings: Option<Earnings>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct ExpirationImpliedVolatility {
         #[serde(with = "string_serialize")]
@@ -135,17 +142,45 @@ pub mod market_metrics {
                 .implied_volatility
         }
     }
+
+    #[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
+    #[serde(rename_all = "kebab-case")]
+    pub struct Earnings {
+        pub expected_report_date: NaiveDate,
+        pub estimated: bool,
+        pub time_of_day: Option<EarningsTimeOfDay>,
+    }
+
+    impl PartialOrd for Earnings {
+        fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
+            Some(self.cmp(other))
+        }
+    }
+
+    impl Ord for Earnings {
+        fn cmp(&self, other: &Self) -> Ordering {
+            self.expected_report_date
+                .cmp(&other.expected_report_date)
+                .then(self.time_of_day.cmp(&other.time_of_day))
+        }
+    }
+
+    #[derive(Clone, Copy, Debug, Eq, Ord, PartialEq, PartialOrd, Serialize, Deserialize, Hash)]
+    pub enum EarningsTimeOfDay {
+        BTO,
+        AMC,
+    }
 }
 
 pub mod positions {
     use super::*;
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub(crate) struct Response {
         pub items: Vec<Item>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct Item {
         pub symbol: String,
@@ -158,7 +193,7 @@ pub mod positions {
         pub instrument_type: String,
     }
 
-    #[derive(Debug, Eq, PartialEq, Serialize, Deserialize)]
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Serialize, Deserialize, Hash)]
     pub enum QuantityDirection {
         Short,
         Long,
@@ -216,17 +251,19 @@ pub mod positions {
 pub mod transactions {
     use super::*;
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub(crate) struct Response {
         pub items: Vec<Item>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
-    #[serde(untagged)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
+    #[serde(tag = "transaction-type")]
     pub enum Item {
         Trade(TradeItem),
+        #[serde(rename = "Receive Deliver")]
         ReceiveDeliver(ReceiveDeliverItem),
-        Other(OtherItem),
+        #[serde(rename = "Money Movement")]
+        MoneyMovement(MoneyMovementItem),
     }
 
     impl Item {
@@ -234,7 +271,7 @@ pub mod transactions {
             match self {
                 Self::Trade(item) => &mut item.id,
                 Self::ReceiveDeliver(item) => &mut item.id,
-                Self::Other(item) => &mut item.id,
+                Self::MoneyMovement(item) => &mut item.id,
             }
         }
 
@@ -242,18 +279,17 @@ pub mod transactions {
             match self {
                 Self::Trade(item) => item.executed_at,
                 Self::ReceiveDeliver(item) => item.executed_at,
-                Self::Other(item) => item.executed_at,
+                Self::MoneyMovement(item) => item.executed_at,
             }
         }
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct TradeItem {
         pub id: u32,
         pub symbol: String,
-        pub instrument_type: String,
-        pub transaction_type: String,
+        pub instrument_type: InstrumentType,
         #[serde(with = "string_serialize")]
         pub executed_at: DateTime<FixedOffset>,
         pub action: TradeAction,
@@ -319,24 +355,54 @@ pub mod transactions {
         }
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct ReceiveDeliverItem {
         pub id: u32,
         pub symbol: String,
-        pub instrument_type: String,
-        pub transaction_type: String,
-        pub transaction_sub_type: String,
+        pub instrument_type: InstrumentType,
+        pub transaction_sub_type: ReceiveDeliverTransactionSubType,
         #[serde(with = "string_serialize")]
         pub executed_at: DateTime<FixedOffset>,
         #[serde(default)]
-        pub action: Option<TradeAction>, // defined for stock splits, missing for exercise
+        // defined for splits, symbols changes and STO/BTO/STC/BTC transaction sub types
+        pub action: Option<TradeAction>,
         pub underlying_symbol: String,
         #[serde(with = "string_serialize")]
         value: Decimal,
-        pub value_effect: ValueEffect,
+        value_effect: ValueEffect,
         #[serde(with = "string_serialize")]
         pub quantity: Decimal,
+        #[serde(default, with = "optional_string_serialize")]
+        clearing_fees: Option<Decimal>,
+        clearing_fees_effect: Option<ValueEffect>,
+        #[serde(default, with = "optional_string_serialize")]
+        regulatory_fees: Option<Decimal>,
+        regulatory_fees_effect: Option<ValueEffect>,
+        #[serde(default, with = "optional_string_serialize")]
+        proprietary_index_option_fees: Option<Decimal>,
+        proprietary_index_option_fees_effect: Option<ValueEffect>,
+    }
+
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Serialize, Deserialize, Hash)]
+    pub enum ReceiveDeliverTransactionSubType {
+        Exercise,
+        Expiration,
+        Assignment,
+        #[serde(rename = "Forward Split")]
+        ForwardSplit,
+        #[serde(rename = "Backwards Split")]
+        BackwardsSplit,
+        #[serde(rename = "Symbol Change")]
+        SymbolChange,
+        #[serde(rename = "Sell to Open")]
+        SellToOpen,
+        #[serde(rename = "Buy to Open")]
+        BuyToOpen,
+        #[serde(rename = "Sell to Close")]
+        SellToClose,
+        #[serde(rename = "Buy to Close")]
+        BuyToClose,
     }
 
     impl PartialEq for ReceiveDeliverItem {
@@ -351,6 +417,20 @@ pub mod transactions {
             self.value_effect.apply(self.value.0)
         }
 
+        pub fn fees(&self) -> Rational64 {
+            self.clearing_fees_effect
+                .map(|v| v.apply(self.clearing_fees.unwrap().0))
+                .unwrap_or_else(|| Rational64::zero())
+                + self
+                    .regulatory_fees_effect
+                    .map(|v| v.apply(self.regulatory_fees.unwrap().0))
+                    .unwrap_or_else(|| Rational64::zero())
+                + self
+                    .proprietary_index_option_fees_effect
+                    .map(|v| v.apply(self.proprietary_index_option_fees.unwrap().0))
+                    .unwrap_or_else(|| Rational64::zero())
+        }
+
         pub fn expiration_date(&self) -> ExpirationDate {
             OptionSymbol::from(&self.symbol).expiration_date()
         }
@@ -368,16 +448,31 @@ pub mod transactions {
         }
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
-    pub struct OtherItem {
+    pub struct MoneyMovementItem {
         pub id: u32,
-        pub transaction_type: String,
         #[serde(with = "string_serialize")]
         pub executed_at: DateTime<FixedOffset>,
+        #[serde(with = "string_serialize")]
+        value: Decimal,
+        value_effect: ValueEffect,
     }
 
-    #[derive(Copy, Clone, Debug, Serialize, Deserialize, Eq, Hash, PartialEq)]
+    impl PartialEq for MoneyMovementItem {
+        fn eq(&self, other: &Self) -> bool {
+            self.id == other.id
+        }
+    }
+    impl Eq for MoneyMovementItem {}
+
+    impl MoneyMovementItem {
+        pub fn value(&self) -> Rational64 {
+            self.value_effect.apply(self.value.0)
+        }
+    }
+
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Serialize, Deserialize, Hash)]
     pub enum TradeAction {
         Sell,
         Buy,
@@ -419,7 +514,7 @@ pub mod transactions {
         }
     }
 
-    #[derive(Copy, Clone, Debug, Serialize, Deserialize, Eq, Hash, PartialEq)]
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Serialize, Deserialize, Hash)]
     pub enum ValueEffect {
         None,
         Debit,
@@ -449,7 +544,12 @@ pub mod transactions {
     impl From<csv::Transaction> for Item {
         fn from(csv: csv::Transaction) -> Self {
             let symbol = csv.symbol.clone().unwrap_or_default();
-            let instrument_type = csv.instrument_type.clone().unwrap_or_default();
+            let instrument_type = serde_json::from_str(
+                csv.instrument_type
+                    .as_ref()
+                    .expect("Missing instrument type"),
+            )
+            .unwrap();
             let underlying_symbol = csv.underlying_symbol().unwrap_or_default().to_string();
 
             let split_fees = Decimal(csv.fees.abs().0 / 3);
@@ -461,7 +561,6 @@ pub mod transactions {
                     id: 0,
                     symbol,
                     instrument_type,
-                    transaction_type: csv.trade_type,
                     executed_at: csv.date,
                     action: csv.action.expect("Missing trade action").into(),
                     underlying_symbol,
@@ -481,24 +580,33 @@ pub mod transactions {
             } else if csv.trade_type == "Receive Deliver" {
                 let description = csv.description.to_ascii_lowercase();
                 let transaction_sub_type = if description.contains("exercise") {
-                    "Exercise".to_string()
+                    ReceiveDeliverTransactionSubType::Exercise
                 } else if description.contains("expiration") {
-                    "Expiration".to_string()
+                    ReceiveDeliverTransactionSubType::Expiration
                 } else if description.contains("assignment") {
-                    "Assignment".to_string()
+                    ReceiveDeliverTransactionSubType::Assignment
                 } else if description.contains("forward split") {
-                    "Forward Split".to_string()
+                    ReceiveDeliverTransactionSubType::ForwardSplit
                 } else if description.contains("backwards split") {
-                    "Backwards Split".to_string()
+                    ReceiveDeliverTransactionSubType::BackwardsSplit
+                } else if description.contains("symbol change") {
+                    ReceiveDeliverTransactionSubType::SymbolChange
+                } else if description.contains("sell to open") {
+                    ReceiveDeliverTransactionSubType::SellToOpen
+                } else if description.contains("buy to open") {
+                    ReceiveDeliverTransactionSubType::BuyToOpen
+                } else if description.contains("sell to close") {
+                    ReceiveDeliverTransactionSubType::SellToClose
+                } else if description.contains("buy to close") {
+                    ReceiveDeliverTransactionSubType::BuyToClose
                 } else {
-                    description
+                    unreachable!("Unhandled transaction sub-type: {}", description);
                 };
 
                 Item::ReceiveDeliver(ReceiveDeliverItem {
                     id: 0,
                     symbol,
                     instrument_type,
-                    transaction_type: csv.trade_type,
                     transaction_sub_type,
                     executed_at: csv.date,
                     action: csv.action.map(|action| action.into()),
@@ -506,13 +614,22 @@ pub mod transactions {
                     value: csv.value.abs(),
                     value_effect: ValueEffect::from_value(csv.value.0),
                     quantity: csv.quantity,
+                    clearing_fees: Some(split_fees),
+                    clearing_fees_effect: Some(fees_effect),
+                    regulatory_fees: Some(split_fees),
+                    regulatory_fees_effect: Some(fees_effect),
+                    proprietary_index_option_fees: Some(split_fees),
+                    proprietary_index_option_fees_effect: Some(fees_effect),
+                })
+            } else if csv.trade_type == "Money Movement" {
+                Item::MoneyMovement(MoneyMovementItem {
+                    id: 0,
+                    executed_at: csv.date,
+                    value: csv.value.abs(),
+                    value_effect: ValueEffect::from_value(csv.value.0),
                 })
             } else {
-                Item::Other(OtherItem {
-                    id: 0,
-                    transaction_type: csv.trade_type,
-                    executed_at: csv.date,
-                })
+                unreachable!("Unhandled transaction type: {}", csv.trade_type);
             }
         }
     }
@@ -532,12 +649,12 @@ pub mod transactions {
 pub mod option_chains {
     use super::*;
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     pub(crate) struct Response {
         pub items: Vec<Item>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct Item {
         pub underlying_symbol: String,
@@ -548,7 +665,7 @@ pub mod option_chains {
         pub expirations: Vec<Expiration>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct TickSize {
         #[serde(with = "string_serialize")]
@@ -557,7 +674,7 @@ pub mod option_chains {
         pub threshold: Option<Decimal>,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct Deliverable {
         pub symbol: Option<String>,
@@ -571,7 +688,7 @@ pub mod option_chains {
         pub percent: i32,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct Expiration {
         pub expiration_type: ExpirationType,
@@ -583,14 +700,14 @@ pub mod option_chains {
         pub strikes: Vec<ExpirationStrike>,
     }
 
-    #[derive(Debug, Eq, PartialEq, Serialize, Deserialize)]
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Serialize, Deserialize, Hash)]
     pub enum ExpirationType {
         Regular,
         Weekly,
         Quarterly,
     }
 
-    #[derive(Debug, Serialize, Deserialize)]
+    #[derive(Clone, Debug, Serialize, Deserialize)]
     #[serde(rename_all = "kebab-case")]
     pub struct ExpirationStrike {
         #[serde(with = "string_serialize")]
